# -*- coding: utf-8 -*-
import logging


_logger = logging.getLogger(__name__)


# https://publicsuffix.org/list/public_suffix_list.dat
PUBLIC_SUFFIX_PATH = '/usr/share/publicsuffix/public_suffix_list.dat'


def canonicalize_label(label):
    if label.isascii():
        return label.lower()
    return encode_idn(label)

def canonicalize(domain):
    return '.'.join(map(canonicalize_label, domain.split('.')))

def encode_idn(domain):
    # bücher.example -> xn--bcher-kva.example
    return domain.encode('idna').decode('ascii')

# def decode_idn(domain):
#     # xn--bcher-kva.example -> bücher.example
#     return domain.encode('utf-8').decode('idna')


class PSLRule(object):
    def __init__(self, line):
        self._raw = line
        self._domain = canonicalize(line)
        self._labels = self._domain.rsplit('.')
        self._label_count = len(self._labels)

    def match(self, domain):
        # A domain is said to match a rule if and only if all of the following conditions are met:
        # - When the domain and rule are split into corresponding labels, that the domain contains
        #   as many or more labels than the rule.
        # - Beginning with the right-most labels of both the domain and the rule, and continuing for
        #   all labels in the rule, one finds that for every pair, either they are identical, or that
        #   the label from the rule is "*".
        domain = canonicalize(domain)

        dlabels = domain.split('.')
        if not len(dlabels) >= self._label_count:
            return False

        j = len(dlabels) - 1
        for i in range(len(self._labels), 0, -1):
            rule_label = self._labels[i - 1]
            domain_label = dlabels[j]
            if not(rule_label == domain_label or rule_label == '*' or rule_label == f'!{domain_label}'):
                return False

            j -= 1

        return True


class PublicSuffixList(object):

    def __init__(self):
        content = None

        try:
            with open(PUBLIC_SUFFIX_PATH, 'r', encoding='utf-8') as f:
                content = f.read()

            self._parse(content)
        except FileNotFoundError:
            _logger.warning('%s not found. Download the file at https://publicsuffix.org/list/public_suffix_list.dat.', PUBLIC_SUFFIX_PATH)

    def _parse(self, content):
        rules = []

        # The Public Suffix List consists of a series of lines, separated by \n.
        for line in content.splitlines():
            if len(line) == 0:
                continue

            # Each line is only read up to the first whitespace; entire lines can also be commented using //.
            line = line.rstrip()
            if line.startswith('//'):
                continue
            line = line.split(' ')[0]
            if line == '':
                continue

            rule = PSLRule(line)
            rules.append(rule)

        self._rules = rules

    def sld(self, domain):
        """ Returns the second-level domain (SLD)."""
        if not isinstance(domain, str):
            return None

        if domain.startswith('.'):
            return None

        dlabels = domain.rsplit('.')

        if len(dlabels) <= 1:
            return None

        matches = []

        # 1. Match domain against all rules and take note of the matching ones.
        for rule in self._rules:
            if rule.match(domain):
                matches.append(rule)

        prevail = None
        # 2. If no rules match, the prevailing rule is "*".
        if len(matches) == 0:
            # Accept unknown TLDs.
            if len(dlabels) >= 2:
                return '.'.join(dlabels[-2:])
            return None

        # 3. If more than one rule matches, the prevailing rule is the one which is an exception rule.
        elif len(matches) > 1:
            exceptions = [m for m in matches if m._domain[0] == '!']

            # 4. If there is no matching exception rule, the prevailing rule is the one with the most labels.
            if not exceptions:
                most = 0
                keep = None
                for m in matches:
                    if m._label_count > most:
                        most = m._label_count
                        keep = m
                prevail = keep

            # 5. If the prevailing rule is a exception rule, modify it by removing the leftmost label.
            else:
                assert len(exceptions) == 1
                r = exceptions[0]
                prevail = PSLRule(r._raw[r._raw.index('.') + 1:])

        else:
            prevail = matches[0]

        # 6. The public suffix is the set of labels from the domain which match the labels of the prevailing rule, using the matching algorithm above.
        j = len(dlabels) - 1
        labels = []
        for i in range(len(prevail._labels), 0, -1):
            if canonicalize_label(prevail._labels[i - 1]) == canonicalize_label(dlabels[j]) or prevail._labels[i - 1] == '*':
                labels.append(dlabels[j].lower())
            j -= 1

        publicsuffix = '.'.join(reversed(labels))
        if domain == publicsuffix:
            return None

        # 7. The registered or registrable domain is the public suffix plus one additional label.
        psd = f'{dlabels[j]}.{publicsuffix}'
        return psd

    def extract(self, hostname):
        sld = self.sld(hostname)
        if sld is None:
            return None

        labels = hostname.split('.')
        llabels = len(labels)
        sld_labels = sld.split('.')
        lpsd = len(sld_labels)

        psd_index = llabels - lpsd
        domain = labels[psd_index]
        subdomain = '.'.join(labels[:psd_index])
        etld = '.'.join(sld_labels[1:])

        return (subdomain, domain, etld)
